
/***************************************************/
/*           D R A G O N   C O N V E R T           */
/*   WAV/VOC/BIN/BAS/DGN/CAS to CAS/DGN/BAS/WAV    */
/*           (c) Sarah Burgin 12/7/1993            */
/* All rights reserved. Commercial use prohibited. */
/***************************************************/

#include "stdio.h"
#include "alloc.h"
#include "ctype.h"
#include "conio.h"
#include "setjmp.h"
#include "bios.h"
#include "string.h"
#include "dir.h"
#include "stdlib.h"

#include "..\lfn\lfn.h"

#define MAJOR			4
#define MINOR			1

#define SENS			30
#define MAX_FILE_SIZE	0xfff0L /* 64K */

#define FILENAME_SIZE	LFN_MAXPATH

#define FAR			    huge
#define FARMALLOC		farmalloc
#define FARFREE 		farfree

#define DGN_ID			0xa1
#define SIG1			0xa2
#define SIG2			0x50

#define FROM_DOS		1
#define FROM_DGN		2
#define FROM_PURE		3
#define FROM_ASCII		4

typedef enum { FALSE, TRUE } boolean;

unsigned char			sens = SENS;
unsigned long			max_file_size = MAX_FILE_SIZE;

FILE            		*fp = NULL;
unsigned char 			FAR *in_file = NULL;
unsigned char 			FAR *address, FAR *block_data, FAR *good_address;
unsigned long			out_address = 0;
signed int				ch;
unsigned char			ch2;
unsigned int			temp, voc_temp, average = 0, wavelength, values;
boolean					voc = FALSE, cas = FALSE, to_wav = TRUE;
boolean					eleven = FALSE, twenty_two = FALSE;
boolean					raw = FALSE, stop_on_error = TRUE;
unsigned long			voc_left;
jmp_buf					jump;
unsigned int			ctrl_chk = 0;

int						intersync = 1;
int						block_number = 0;
boolean					block_number_on = FALSE;
boolean					first_file = TRUE;
boolean					ascii_basic = FALSE;
fpos_t					good_fpos;
signed long				temp_val;
boolean					do_leader = TRUE;
boolean					block_error;
boolean					allow_ascii = TRUE;

const unsigned char		valid_wav[] = "WAVEfmt";
const unsigned char		valid_voc[] = "Creative Voice File";

unsigned char			block_type = 0, block_len, block_check;

unsigned char			far	*mem, far *mem2;
unsigned char			*ffname, *namebase;
long					pointer = 0, outptr = 0;
unsigned char			count, count2;
unsigned long			cas_address, output_size;
unsigned char			output;
unsigned char			chk;
char					source = FROM_PURE;

unsigned int			start = 0x600, exec = 359, length;
unsigned char			type = 0x02, ascii_flag = 0, gap_flag = 0;

#define SINEWAVE_ELEMENTS11 10
const unsigned char sinewave_data11[]
	= { 0x83,     0xC4,     0xF5,
			 0xF3,     0xC2,
		0x77,     0x34,     0x06,
			 0x0D,     0x42      };

#define SINEWAVE_ELEMENTS22 20
const unsigned char sinewave_data22[]
	= { 0x83,0xA5,0xC4,0xE2,0xF5,
		0xFB,0xF3,0xDE,0xC2,0xA0,
		0x77,0x51,0x34,0x16,0x06,
		0x03,0x0D,0x25,0x42,0x65 };

#define SINEWAVE_ELEMENTS44 42
const unsigned char sinewave_data44[]
	= {	0x83,0x8F,0xA2,0xB3,0xC1,
		0xCF,0xDC,0xEA,0xF2,0xF9,
		0xFB,0xFB,0xF9,0xF2,0xEB,
		0xDE,0xD2,0xC4,0xB6,0xA6,
		0x92,0x7E,0x70,0x5D,0x4C,
		0x3E,0x31,0x23,0x15,0x0D,
		0x06,0x03,0x03,0x04,0x0B,
		0x12,0x1D,0x2B,0x39,0x47,
		0x55,0x69 };

const unsigned char header1[] = {0x52, 0x49, 0x46, 0x46};

const unsigned char header11[] =
	{
		'W','A','V','E','f','m','t',32,
		0x10,0,0,0,
		1,0,1,0,
		0x11, 0x2b, 0 ,0,
		0x11, 0x2b, 0 ,0,
		1,0,8,0,
		0x64, 0x61, 0x74, 0x61
	};

const unsigned char header22[] =
	{
		'W','A','V','E','f','m','t',32,
		0x10,0,0,0,
		1,0,1,0,
		0x22, 0x56, 0 ,0,
		0x22, 0x56, 0 ,0,
		1,0,8,0,
		0x64, 0x61, 0x74, 0x61
	};

const unsigned char header44[] =
	{
		'W','A','V','E','f','m','t',32,
		0x10,0,0,0,
		1,0,1,0,
		0x44, 0xac, 0 ,0,
		0x44, 0xac, 0 ,0,
		1,0,8,0,
		0x64, 0x61, 0x74, 0x61
	};

void cerror(const unsigned char *eout)
{
	fprintf(stderr, eout);
	exit(0);
}

FILE *fopenout(const char *outname)
{
	char input = 0;

	if (lfn_access(outname, 0) == 0)
	{
		printf("Overwrite %s (Y/N)? ", outname);
		while ((input != 'Y') && (input != 'N'))
		{
			switch (input = toupper(getch()))
			{
				case 'N':	fputchar('N');
				case 27:
				case 3:		cerror("\nAborted!");
				case 'Y':	printf("Y\n");
			}
		}
	}
	return(lfn_fopen(outname, "wb"));
}

void put_header1(void)
{
	for (count2 = 0; count2 < 4; count2++)
		fputc(header1[count2],fp);
}

void put_header2(void)
{
	for (count2 = 0; count2 < 32; count2++)
	{
		if (eleven)
			fputc(header11[count2],fp);
		else if (twenty_two)
			fputc(header22[count2],fp);
		else
			fputc(header44[count2],fp);
	}
}

void put_hex(unsigned long hex_input)
{
	fputc(hex_input         & 0xff, fp);
	fputc((hex_input >> 8)  & 0xff, fp);
	fputc((hex_input >> 16) & 0xff, fp);
	fputc((hex_input >> 24) & 0xff, fp);
}

void put_bit(unsigned char bbit)
{
	if (eleven)
	{
		if (output)
		{
			count2 = 0;
			while (count2 < SINEWAVE_ELEMENTS11)
			{
				fputc((sinewave_data11[count2++]), fp);
				count2 += bbit;
			}
			fputc((sinewave_data11[(count2 - bbit) -1]), fp);
		}
		else
			output_size += ((SINEWAVE_ELEMENTS11 >> bbit) + 1);
	}
	else if (twenty_two)
	{
		if (output)
		{
			count2 = 0;
			while (count2 < SINEWAVE_ELEMENTS22)
			{
				fputc((sinewave_data22[count2++]), fp);
				count2 += bbit;
			}
			fputc((sinewave_data22[(count2 - bbit) -1]), fp);
		}
		else
			output_size += ((SINEWAVE_ELEMENTS22 >> bbit) + 1);
	}
	else
	{
		if (output)
		{
			count2 = 0;
			while (count2 < SINEWAVE_ELEMENTS44)
			{
				fputc((sinewave_data44[count2++]), fp);
				count2 += bbit;
			}
			fputc((sinewave_data44[(count2 - bbit) -1]), fp);
		}
		else
			output_size += ((SINEWAVE_ELEMENTS44 >> bbit) + 1);
	}
}

void put_byte(void)
{
	count = 8;
	while (count > 0)
	{
		put_bit(ch & 0x01);
		ch >>= 1;
		count--;
	}
}

void cas2wav(unsigned char *fn1)
{
	strcat(fn1,".wav");

	printf("=\n");
	if (raw || (!do_leader) || (average != 0) || (sens != SENS) || (!allow_ascii))
		cerror("Option not applicable for cas to wav conversions.");

	output = FALSE;
	ch = output_size = cas_address = 0;
	while (ch != EOF)
	{
		if (cas_address >= max_file_size)
			cerror("File is too big!");
		in_file[cas_address++] = ch = fgetc(fp);
		put_byte();
	}
	fclose(fp);

	printf("Sample rate ");
	if (eleven)
		printf("11025Hz");
	else if (twenty_two)
		printf("22050Hz");
	else
		printf("44100Hz");
	printf("; output size %uK.\n", (output_size / 1024));

	if ((fp = fopenout(fn1)) == NULL)
		cerror("Cannot open output file.");
	printf("Writing %s 00%%",fn1);

	/* Turn file output on. */
	output = TRUE;

	/* Write header. */
	put_header1();
	put_hex(output_size + 0x24);
	put_header2();
	put_hex(output_size);

	output_size = 0;
	while (output_size < cas_address)
	{
		ch = in_file[output_size++];
		put_byte();
		if ((output_size % 64) == 0)
		{
			printf("%c%c%c%02lu%%",
				8,8,8,(100 * (unsigned long)output_size) / (unsigned long)cas_address);

			if (bioskey(2) & 0x0004)
			{
				fputchar('\n');
				fclose(fp);
				remove(fn1);
				cerror("Aborted!");
			}
		}
	}
	fclose(fp);
	printf("%c%c%c   ",8,8,8);
	exit(0);
}

unsigned char doc(unsigned char chin)
{
	if ((chin >= '#') && (chin <= ')'))
		return(chin);
	if ((chin >= '0') && (chin <= '9'))
		return(chin);
	chin = toupper(chin);
	if ((chin >= '@') && (chin <= 'Z'))
		return(chin);
	return('_');
}

void dos(void)
{
	type = fgetc(fp);
	if (type == 0x01)
	{
		printf("=BAS=\n");
		type = 0;
	}
	else
	{
		printf("=BIN=\n");
	}

	start = (fgetc(fp) << 8);
	start += fgetc(fp);

	length = (fgetc(fp) << 8);
	length += fgetc(fp);

	exec = (fgetc(fp) << 8);
	exec += fgetc(fp);

	source = FROM_DOS;
}

void dgn(void)
{
	/* Get filename. */
	for (count = 0; count < 8; count++)
		namebase[count] = fgetc(fp);

	type = fgetc(fp);
	ascii_flag = fgetc(fp);
	gap_flag = fgetc(fp);

	exec = (fgetc(fp) << 8);
	exec += fgetc(fp);

	start = (fgetc(fp) << 8);
	start += fgetc(fp);

	source = FROM_DGN;
}

void out(unsigned char outc)
{
	chk += outc;
	fputc(outc,fp);
}

void other_2cas(char *openrawname)
{
	ffname		= (unsigned char *)malloc(FILENAME_SIZE);
	namebase	= (unsigned char *)malloc(FILENAME_SIZE);
	mem			= (unsigned char far *)farmalloc(MAX_FILE_SIZE);

	if ((ffname == NULL) || (namebase == NULL) || (mem == NULL))
		cerror("Failed to re-allocate memory.");

	/* Rewind input file. */
	rewind(fp);

	/* Get filename without path. */
	strcpy(namebase, openrawname);
	while (strchr(namebase, '\\') != NULL)
		strcpy(namebase,strchr(namebase, '\\') + 1);
	while (strchr(namebase, '/') != NULL)
		strcpy(namebase,strchr(namebase, '/') + 1);

	/* Establish file type. */
	switch (ch = fgetc(fp))
	{
		case 0x55:	dos();
					break;
		case 0xa1:	printf("=DGN=\n");
					dgn();
					break;
		case EOF:	cerror("Zero length file!");
		default:	mem[pointer++] = ch;
					break;
	}

	if ((ch = fgetc(fp)) == EOF)
		cerror("Empty file body!");

	mem[pointer++] = ch;

	do
	{
		if((ch = fgetc(fp)) != EOF)
			mem[pointer++] = ch;
	} while ((ch != EOF) && (pointer < (MAX_FILE_SIZE-255)));

	if (pointer == (MAX_FILE_SIZE-255))
		cerror("File is too big!");

	fclose(fp);

	if (source == FROM_DOS)
		pointer = length;

	if (source == FROM_PURE)
	{
		long curchar, bincount;

		for (curchar = bincount = 0; curchar < pointer; curchar++)
		{
			if (mem[curchar] > 0x7f)
				bincount++;
		}

		if ((((bincount * 100) / pointer) < 5) && (allow_ascii))
		{
			boolean crlf, donumber;

			printf("=ASCII=\n");

			source = FROM_ASCII;
			type = 0;
			gap_flag = ascii_flag = 0xff;

			mem2 = (unsigned char far *)farmalloc(MAX_FILE_SIZE);
			if (mem2 == NULL)
				cerror("Too little memory to perform conversion!");

			bincount = 0;
			mem2[bincount++] = 0x0d;
			crlf = TRUE;
			for (curchar = 0; curchar < pointer; curchar++)
			{
				ch = mem[curchar];
				if ((ch == 0x0d) || (ch == 0x0a))
				{
					if (!crlf)
					{
						crlf = TRUE;
						mem2[bincount++] = 0x0d;
					}
				}
				else
				{
					if (crlf)
						donumber = TRUE;
					if (donumber)
					{
						if ((ch < '0') || (ch > '9'))
						{
							if (ch != ' ')
							{
								donumber = FALSE;
								if (!crlf)
									mem2[bincount++] = ' ';
							}
						}
						if (ch != ' ')
							mem2[bincount++] = ch;
					}
					else
						mem2[bincount++] = ch;
					crlf = FALSE;
				}
			}
			if (!crlf)
				mem2[bincount++] = 0x0d;
			pointer = bincount;
			farfree(mem);
			mem = mem2;
		}
		else
			printf("=BINARY=\n");
	}
	else if (!allow_ascii)
	{
		cerror("Binary option only available for ASCII input files.");
	}

	if (raw || (!do_leader) || (average != 0) || (sens != SENS)
			|| (eleven) || (twenty_two))
		cerror("Option not applicable to this file type.");

	if ( ((type != 0x00) && (type != 0x02))
			|| ((type == 0x02) && ((ascii_flag + gap_flag) > 0))
			|| ((type == 0x00) && (source != FROM_ASCII)
				&& ((ascii_flag + gap_flag) > 0)))
		printf("Warning: Unknown internal type which may not be converted correctly!\n");

	strcat(namebase,"       ");
	for (count = 0; count < 8; count++)
	{
		ffname[count] = doc(namebase[count]);
	}
	if (ffname[7] == '_')
	{
		ffname[7] = 0;
		if (ffname[6] == '_')
		{
			ffname[6] = 0;
			if (ffname[5] == '_')
			{
				ffname[5] = 0;
				if (ffname[4] == '_')
				{
					ffname[4] = 0;
					if (ffname[3] == '_')
					{
						ffname[3] = 0;
						if (ffname[2] == '_')
						{
							ffname[2] = 0;
							if (ffname[1] == '_')
							{
								ffname[1] = 0;
							}
						}
					}
				}
			}
		}
	}
	else
		ffname[8] = 0;

	if (strcmp(ffname,"_") == NULL)
		strcpy(ffname,"cassette");

	strcat(ffname,".cas");

	if ((fp = fopenout(ffname)) == NULL)
		cerror("Cannot open output file.");
	printf("Writing %s",ffname);

	for (count = 0; count < 16; count++)
		fputc(0x55,fp);

	fputc(0x3c,fp);
	chk=0;
	out(0);
	out(20);

	for (count = 0; count < 8; count++)
		out(toupper(namebase[count]));

	out(type);
	out(ascii_flag);
	out(gap_flag);
	out((exec >> 8) & 0x00ff);
	out(exec & 0x00ff);
	out((start >> 8) & 0x00ff);
	out(start & 0x00ff);
	out(SIG1);
	out(SIG2);
	out(MAJOR);
	out(MINOR);
	out(source);
	fputc(chk,fp);

	for (count = 0; count < 7; count++)
		fputc(0x55,fp);

	while (outptr < pointer)
	{
		fputc(0x55,fp);
		fputc(0x3c,fp);
		chk=0;
		out(0x01);

		if ((pointer - outptr) < 255L)
			count = (pointer-outptr);
		else
			count = 255;

		out(count);

		while(count>0)
		{
			out(mem[outptr++]);
			count--;
		}

		fputc(chk,fp);
		fputc(0x55,fp);
	}

	fputc(0x55,fp);
	fputc(0x3c,fp);
	chk=0;
	out(0xff);
	out(0x00);
	fputc(chk,fp);
	fputc(0x55,fp);

	fclose(fp);
	exit(0);
}

void error(char fatal, const unsigned char *eout)
{
	if (block_number_on)
	{
		fputchar('\n');
		block_number_on = FALSE;
	}
	if (fatal)
		printf("Fatal");
	else
		printf("?IO");
	printf(" Error: %s",eout);
	if (fatal)
		exit(1);
	else
	{
		if (stop_on_error)
		{
			unsigned char fname[FILENAME_SIZE];
			char input = 0;

			printf("\nAbort, Retry, Proceed or Ignore all? ");
			while ((input != 'P') && (input != 'I') && (input != 'R'))
			{
				switch (input = toupper(getch()))
				{
					case 'A':	fputchar('A');
					case 27:
					case 3:		cerror("\nAborted!");
					case 'I':	stop_on_error = FALSE;
					case 'P':	break;
					case 'R':	printf("R\n");
								address = good_address;
								fsetpos(fp, &good_fpos);
								block_error = TRUE;
								temp_val = 0;
								while (temp_val == 0)
								{
									printf("Wave length [%02d]: ",average);
									gets(fname);
									sscanf(fname, "%lu", &temp_val);
									if ((temp_val >= 1) && (temp_val <= 300))
										average = temp_val;
									else
									{
										if (fname[0] == '\0')
											temp_val = 1;
										else
										{
											printf("Wavelength threshold must be in range 1 to 300.\n");
											temp_val = 0;
										}
									}
								}
								temp_val = 0;
								while (temp_val == 0)
								{
									printf("Wave volume [%02d]: ", sens);
									gets(fname);
									sscanf(fname, "%lu", &temp_val);
									if ((temp_val >= 1) && (temp_val <= 126))
										sens = temp_val;
									else
									{
										if (fname[0] == '\0')
											temp_val = 1;
										else
										{
											printf("Input volume must be in range 1 to 126.\n");
											temp_val = 0;
										}
									}
								}
								break;
				}
			}
			if (input != 'R')
				printf("%c\n",input);
		}
		else
			fputchar('\n');
	}
}

void get_ch(void)
{
	if ((ch = fgetc(fp)) == EOF)
	{
		if (cas)
			error(1,"Reached end of file.");
		longjmp(jump, 1);
	}
	if ((++ctrl_chk % 2048) == 0)
	{
		if (bioskey(2) & 0x0004)
		{
			if (block_number_on)
				fputchar('\n');
			cerror("Aborted!");
		}
	}
}

unsigned char get_char(void)
{
	unsigned char block_type;

	get_ch();
	if (voc)
	{
		if (voc_left < 1)
		{
			while ((ch != 0x01) && (ch != 0x02))
			{
				switch (ch)
				{
					/* All loops here have 1 iteration more than just      */
					/* skipping the unwanted data; to get next block type. */

					case 0x00:	longjmp(jump, 1);

					case 0x01:
					case 0x02:	break;

					case 0x03:	for (voc_temp = 7; voc_temp > 0; voc_temp--)
									get_ch();
								break;

					case 0x04:	for (voc_temp = 6; voc_temp > 0; voc_temp--)
									get_ch();
								break;

					case 0x05:	get_ch();
								voc_temp = ch + 1;
								get_ch();
								voc_temp += (ch << 8);
								get_ch();
								voc_temp += (ch << 16);
								while (voc_temp > 0)
								{
									get_ch();
									voc_temp--;
								}
								break;

					case 0x06:	error(1,"Cannot operate on repeated voc data.");

					case 0x07:	error(1,"Misplaced end_repeat in voc data.");

					case 0x08:	for (voc_temp = 8; voc_temp > 0; voc_temp--)
									get_ch();
								break;

					default:	error(1,"Unknown code in voc data.");
				}
			}

			block_type = ch;

			get_ch();
			voc_left = ch;
			get_ch();
			voc_left += ((unsigned)ch << 8);
			get_ch();
			voc_left += ((unsigned long)ch << 16);

			if (block_type == 1)
			{
				voc_left -= 2;
				get_ch();
				get_ch();
				if (ch > 0)
					error(1,"Cannot operate on packed voc data.");
			}
			get_ch();
		}
		voc_left--;
	}
	return(ch);
}

void trace_sinewave(void)
{
	wavelength = 0;

	/* Trace rising edge. */
	while (get_char() <= (0x80 + sens))
		wavelength++;

	/* Trace falling edge. */
	while (get_char() >= (0x80 - sens))
		wavelength++;

	/* Trace 2nd rising edge. */
	while (get_char() <= 0x68)
		wavelength++;
}

unsigned char get_bit(void)
{
	trace_sinewave();
	return(wavelength < average);
}

unsigned char get_byte(void)
{
	if (cas)
	{
		ch2 = get_char();
	}
	else
	{
		ch2 = values = 0;
		while (values < 8)
		{
			ch2 = ((ch2 >> 1) + (get_bit() << 7));
			values++;
		}
		if ((unsigned long)(address - in_file) < max_file_size)
		{
			*address = ch2;
			address = address + 1;
		}
		else
			error(1,"File is too big!");
	}
	return(ch2);
}

void get_sync_byte(unsigned char sbyte)
{
	ch2 = 0;
	while (ch2 != sbyte)
	{
		if (cas)
			ch2 = get_char();
		else
			ch2 = ((ch2 >> 1) + (get_bit() << 7));
	}
	if (!cas)
	{
		if ((unsigned long)(address - in_file) < max_file_size)
		{
			*address = sbyte;
			address = address + 1;
		}
		else
			error(1,"File is too big!");
	}
}

void get_leader_bytes(void)
{
	int leaders = 16-intersync;
	if (leaders < 1)
		leaders = 0;

	for (temp = leaders; temp > 0; temp--)
		get_sync_byte(0x55);
}

void get_block(void)
{
	block_check = 0;
	block_error = FALSE;

	get_sync_byte(0x3c);
	block_check += (block_type = get_byte());
	block_check += (block_len  = get_byte());

	block_data = (unsigned char FAR *)address;
	for (temp = block_len; temp > 0; temp--)
	{
		block_check += get_byte();
		if (cas)
		{
			if ((unsigned long)(address - in_file) < max_file_size)
			{
				*address = ch2;
				address = address + 1;
			}
			else
				error(1,"File is too big!");
		}
	}

	if (block_check != get_byte())
		error(cas,"Block checksum failure.");
	else if ((block_type != 0xff) && (block_type > 1))
		error(cas,"Unknown block type.");
}

char *strlastdot(const char *s)
{
	const char *sret;

	sret = &s[strlen(s)];

	while (sret >= s)
	{
		if (*sret == '.')
			return((char *)sret);
		else if ((*sret == '\\') || (*sret == '/'))
			break;
		else
			sret--;
	}
	return(NULL);
}

void main(int argc, unsigned char *argv[])
{
	unsigned char fname[2][FILENAME_SIZE];
	unsigned char arg_num = 1;

	if (argc < 2)
	{
		printf("\nDRAGON CONVERT V%d.%d (C) S.Burgin %s\n%s%s%s%s%s%s%s%s%s%s%s%s",
			MAJOR, MINOR,
			__DATE__, "Usage: ", argv[0], " [options] [filename[.ext]]\n",
					"\t/Sv\tsample file input volume v (v=1..126)\n",
					"\t/Wt\tsample file wavelength threshold t (t=1..300)\n",
					"\t/L\tdo not assume sample file has an initial leader\n",
					"\t/R\tread sample file as an unformatted bitstream\n",
					"\t/B\tconvert text file to binary cas, not ascii cas\n",
					"\t/D\tconvert cas file to dgn or bas, not wav\n",
					"\t/1\toutput 11.025KHz wav file\n",
					"\t/2\toutput 22.050KHz wav file\n",
					"\t/4\toutput 44.100KHz wav file (default)\n");
		exit(0);
	}

	/* Process command line inputs. */
	while ((argc > 1) && ((argv[arg_num][0] == '/') || (argv[arg_num][0] == '-')))
	{
		switch (toupper(argv[arg_num][1]))
		{
			case 'S':
				temp_val = 0;
				sscanf(&argv[arg_num][2], "%lu", &temp_val);
				if ((temp_val >= 1) && (temp_val <= 126))
					sens = temp_val;
				else
					cerror("Input volume must be in range 1 to 126.");
				break;

			case 'W':
				temp_val = 0;
				sscanf(&argv[arg_num][2], "%lu", &temp_val);
				if ((temp_val >= 1) && (temp_val <= 300))
					average = temp_val;
				else
					cerror("Wavelength threshold must be in range 1 to 300.");
				break;

			case 'R':
				raw = TRUE;
				break;

			case 'L':
				do_leader = FALSE;
				break;

			case 'B':
				allow_ascii = FALSE;
				break;

			case 'D':
				to_wav = FALSE;
				break;

			case '1':
				eleven = TRUE;
				twenty_two = FALSE;
				break;

			case '2':
				eleven = FALSE;
				twenty_two = TRUE;
				break;

			case '4':
				eleven = FALSE;
				twenty_two = FALSE;
				break;

			default:
				sprintf(fname[0],"Unknown option %s.",argv[arg_num]);
				cerror(fname[0]);
		}
		argc--;
		arg_num++;
	}

	if ((raw) && (!do_leader))
		cerror("Leader option is always selected for raw files.");

	if ((in_file = (unsigned char FAR *)FARMALLOC(max_file_size)) == NULL)
		cerror("Cannot allocate memory.");

	if (argc < 2)
		strcpy(fname[0], "dragon");
	else
		strcpy(fname[0], argv[arg_num]);
	strcpy(fname[1], fname[0]);
	if (strlastdot(fname[1]) != NULL)
		*strlastdot(fname[1]) = '\0';

	if ((fp = lfn_fopen(fname[0], "rb")) == NULL)
	{
		sprintf(fname[0],"%s.wav",fname[1]);
		if ((fp = lfn_fopen(fname[0], "rb")) == NULL)
		{
			sprintf(fname[0],"%s.voc",fname[1]);
			if ((fp = lfn_fopen(fname[0], "rb")) == NULL)
			{
				sprintf(fname[0],"%s.bin",fname[1]);
				if ((fp = lfn_fopen(fname[0], "rb")) == NULL)
				{
					sprintf(fname[0],"%s.bas",fname[1]);
					if ((fp = lfn_fopen(fname[0], "rb")) == NULL)
					{
						sprintf(fname[0],"%s.dgn",fname[1]);
						if ((fp = lfn_fopen(fname[0], "rb")) == NULL)
						{
							sprintf(fname[0],"%s.cas",fname[1]);
							if ((fp = lfn_fopen(fname[0], "rb")) == NULL)
							{
								sprintf(fname[0],"Failed to open %s.wav/voc/bin/bas/dgn/cas!",fname[1]);
								cerror(fname[0]);
							}
						}
					}
				}
			}
		}
	}

	/* Identify file type and skip header. */
	get_ch();
	switch(ch)
	{
		case 'R':	for (temp = 7; temp > 0; temp--)
						get_ch();
					for (temp = 7; temp > 0; temp--)
					{
						get_ch();
						if (ch != valid_wav[7-temp])
							cerror("Not a valid .wav file.");
					}
					for(temp = 28; temp > 0; temp--)
						get_ch();
					printf("=WAV");
					break;

		case 'C':	voc = TRUE;
					for (temp = 18; temp > 0; temp--)
					{
						get_ch();
						if (ch != valid_voc[19-temp])
							cerror("Not a valid .voc file.");
					}
					for (temp = 7; temp > 0; temp--)
						get_ch();
					voc_left = 0;
					printf("=VOC");
					break;

		case 0x55:	/* Dragon-DOS binary or .CAS file? */
					for (temp = 8; temp > 0; temp--)
						get_ch();
					if (ch == 0xaa)
					{
						FARFREE(in_file);
						other_2cas(fname[1]);
					}
					rewind(fp);
					get_ch();
					printf("=CAS");
					if (to_wav)
						cas2wav(fname[1]);
					cas = TRUE;
					break;

		default:	FARFREE(in_file);
					other_2cas(fname[1]);
	}
	printf("=\n");

	if (!allow_ascii)
		cerror("Binary option not available for this file type.");

	if (eleven)
		cerror("11Khz option not available for this file type.");
	else if (twenty_two)
		cerror("22Khz option not available for this file type.");

	/* Strip name from output filename stub. */
	if (!cas)
	{
		do
		{
			fname[1][strlen(fname[1])-1] = ch2 = 0;
			if (strlen(fname[1]) > 0)
				ch2 = fname[1][strlen(fname[1])-1];
		} while ((ch2 != '\\') && (ch2 != 0) && (ch2 != '/'));
	}

	/* Calculate average sinewave wavelength. */
	if (!cas)
	{
		if (average == 0)
		{
			/* Seeking first sinewave. */
			if (setjmp(jump) == 0)
			{
				int wavesize;

				/* Ignore initial noise. */
				for (temp = 3; temp > 0; temp--)
				{
					do
					{
						wavesize = 0;
						while (get_char() < (0x80 + sens))
							wavesize++;
						while (get_char() > (0x80 - sens))
							wavesize++;
						while (get_char() < (0x80 + sens + (sens >> 1)))
							wavesize++;
						while (get_char() > (0x80 - sens + (sens >> 1)))
							wavesize++;
						while (get_char() < 0x80)
							wavesize++;
					} while (wavesize < 20);
				}

				/* Calculate average wavelength. */
				values = average = 0;
				for (temp = 40; temp > 0; temp--)
				{
					trace_sinewave();
					values++;
					average += wavelength;
				}
				average = (average / values) + 1;
				printf("Wavelength threshold %d.\n", average);
				rewind(fp);
			}
			else
				cerror("Failed to calculate wavelength threshold.");
		}
	}
	else
	{
		if (!do_leader)
			cerror("Leader option is always selected for .cas files.");
		else if (raw)
			cerror("Raw option has no meaning for .cas files.");
		else
			printf("Extracting from cassette format.\n");
		do_leader = FALSE;
	}

	/* Insert file id byte in buffer. */
	*in_file = (cas) ? DGN_ID : 0x55;

	/* Set load address to buffer start. */
	address = in_file + 1;

	/* Read the data from the input file. */
	if (setjmp(jump) == 0)
	{
		if (raw)
		{
			first_file = FALSE;
			strcat(fname[1],"cassette.cas");
			printf ("Raw file 0000 bytes");
			block_number_on = TRUE;
			for (;;)
			{
				get_byte();
				if (((address-in_file) & 0x00ff) == 0)
				{
					printf("%c%c%c%c%c%c%c%c%c%c%04X bytes",
						8,8,8,8,8,8,8,8,8,8,
						address-in_file);
				}
			}
		}

		/* Synchronize to first section of leader bytes. */
		if (do_leader)
			get_leader_bytes();

		for (;;)
		{

			/* Read next block. */
			if ((cas) && (first_file))
				address = in_file + 1;

			good_address = address;
			(void)fgetpos(fp, &good_fpos);

			get_block();
			if (!block_error)
			{
				block_number++;

				/* Add additional 0x55 characters. */
				if (!cas)
				{
					int syncs = intersync;

					while (syncs > 0)
					{
						syncs--;
						if ((unsigned long)(address - in_file) < max_file_size)
						{
							*address = 0x55;
							address = address + 1;
						}
						else
							error(1,"File is too big!");
					}
				}

				switch (block_type)
				{
				case 0x00:
						if ((cas) && (!first_file))
							longjmp(jump, 1);
						block_number = 0;
						if (block_number_on)
						{
							fputchar('\n');
							block_number_on = FALSE;
						}
						switch(block_data[8])
						{
							case 0:	printf("Header: ");
									if (block_data[9] != 0)
										printf("ASCII");
									else
										printf("Tokenised");
									printf(" BASIC.");
									break;

							case 1:	if (cas)
										cerror("Cannot convert data file to .dgn format.");
									else
										printf("Header: Data file.");
									break;

							case 2: printf("Header: Binary ");
									printf("(start %04X, ",(block_data[13] << 8) + block_data[14]);
									printf("exec %04X).",(block_data[11] << 8) + block_data[12]);
									break;

							default: printf("Header: Unknown type.");
									break;
						}
						fputchar('\n');

						if (first_file)
						{
							first_file = FALSE;

							if (cas)
							{
								if ((block_data[8] == 0)
										&& (block_data[9] != 0))
								{
									strcat(fname[1],".bas");
									ascii_basic = TRUE;
									out_address = 16;
								}
								else
									strcat(fname[1],".dgn");

								address = block_data + 15;
								if (block_len > 15)
									printf("Warning: Header block larger than 15 bytes; extras discarded.\n");
							}
							else
							{
								/* DOS translation of Dragon filename. */
								ch2 = strlen(fname[1]);
								fname[1][ch2    ] = doc(block_data[0]);
								fname[1][ch2 + 1] = doc(block_data[1]);
								fname[1][ch2 + 2] = doc(block_data[2]);
								fname[1][ch2 + 3] = doc(block_data[3]);
								fname[1][ch2 + 4] = doc(block_data[4]);
								fname[1][ch2 + 5] = doc(block_data[5]);
								fname[1][ch2 + 6] = doc(block_data[6]);
								fname[1][ch2 + 7] = doc(block_data[7]);
								if (fname[1][ch2 + 7] == '_')
								{
									fname[1][ch2 + 7] = 0;
									if (fname[1][ch2 + 6] == '_')
									{
										fname[1][ch2 + 6] = 0;
										if (fname[1][ch2 + 5] == '_')
										{
											fname[1][ch2 + 5] = 0;
											if (fname[1][ch2 + 4] == '_')
											{
												fname[1][ch2 + 4] = 0;
												if (fname[1][ch2 + 3] == '_')
												{
													fname[1][ch2 + 3] = 0;
													if (fname[1][ch2 + 2] == '_')
													{
														fname[1][ch2 + 2] = 0;
														if (fname[1][ch2 + 1] == '_')
														{
															fname[1][ch2 + 1] = 0;
														}
													}
												}
											}
										}
									}
								}
								else
									fname[1][ch2 + 8] = 0;

								if (strcmp(&fname[1][ch2],"_") == NULL)
									strcpy(&fname[1][ch2],"cassette");

								strcat(fname[1],".cas");
							}
						}
						for (block_len = 0; block_len < 8; block_len++)
						{
							fname[0][block_len] = block_data[block_len];
							if (!isprint(fname[0][block_len]))
								fname[0][block_len] = '?';
						}
						fname[0][8] = '\0';
						break;

				case 0xff:
						if (cas)
						{
							if (block_number_on)
								printf("%c%c%02X", 8,8, block_number + 1);
							longjmp(jump, 1);
						}
						break;

				default:
						break;

				}
			}

			if (!block_number_on)
			{
				if (first_file)
				{
					if (!block_error)
					{
						strcpy(fname[0],"HEADLESS");
						printf("Loading %s %02X", fname[0], block_number + 1);
						block_number_on = TRUE;
					}
				}
				else
				{
					printf("Loading %s %02X", fname[0], block_number + 1);
					block_number_on = TRUE;
				}
			}
			else
			{
				printf("%c%c%02X", 8,8, block_number + 1);
			}

			if ((block_type == 0xff) && (!block_error))
			{
				block_number = 0;
				fputchar('\n');
				block_number_on = FALSE;
				strcpy(fname[0],"HEADLESS");
			}
		}
	}
	fclose(fp);

	/* Tidy up output. */
	if (block_number_on)
	{
		if (raw)
		{
			/* Print final raw file length. */
			printf("%c%c%c%c%c%c%c%c%c%c%04X bytes.",
				8,8,8,8,8,8,8,8,8,8,
				address-in_file);
		}
		fputchar('\n');
	}

	if (first_file)
	{
		if (cas)
			cerror("Nothing to write!");
		else
			strcat(fname[1],"HEADLESS.cas");
	}

	if ((fp = fopenout(fname[1])) == NULL)
		cerror("Cannot open output file.");
	printf("Writing %s", fname[1]);

	while(out_address < (address - in_file))
	{
		fputc(in_file[out_address++], fp);
		if (ascii_basic && cas && (in_file[out_address - 1] == 0x0d))
			fputc(0x0a, fp);
	}
	fclose(fp);
}
